@@ -1,8 +1,10 @@
 from __future__ import annotations
 
-import collections
 import hashlib
 import re
+from collections import Counter
+from collections import defaultdict
+from copy import copy
 from typing import Callable
 from uuid import uuid4
 
@@ -14,6 +16,30 @@ from pyspark.sql.types import DataType
 from pyspark.sql.types import StructType
 
 
+class _DeconstructedField:
+    def __init__(self, field: str):
+        split_field = field.split('.')
+        self.levels = copy(split_field)
+        self.trunk = split_field.pop(0)
+        if len(split_field) > 1:
+            *self.branches, self.leaf = split_field
+        elif len(split_field) == 0:
+            self.branches, self.leaf = [], None
+        else:
+            self.branches, self.leaf = [], split_field[0]
+        self.trunk_and_branches = '.'.join([self.trunk] + self.branches)
+
+
+def _field_trie(fields: list[str]):
+    result = defaultdict(list)
+    for field in fields:
+        deconstructed_field = _DeconstructedField(field)
+        trunk_and_branches = deconstructed_field.trunk_and_branches
+        leaf = deconstructed_field.leaf
+        result[trunk_and_branches].append(leaf)
+    return result
+
+
 def _fields(
     df: DataFrame,
     include_types: bool,
@@ -85,9 +111,6 @@ def deduplicate_dataframe(
     """
     Removes duplicates from a Spark DataFrame.
 
-    can we validate without keys? not meaningfully
-    is there an instance where we would want to provide keys but also specify full_row dupes?
-
     :param df: The target Delta Lake table that contains duplicates.
     :param keys: A list of column names used to distinguish rows. The order of this list does not matter.
     :param tiebreaking_columns: A list of column names used for ordering. The order of this list matters, with earlier elements "weighing" more than lesser ones. The columns will be evaluated in descending order. In the event of a tie, you will get non-deterministic results.
@@ -99,7 +122,7 @@ def deduplicate_dataframe(
     if tiebreaking_columns is None:
         tiebreaking_columns = []
 
-    if isinstance(keys, str):  # pragma: no cover
+    if isinstance(keys, str):
         keys = [keys]
 
     if not keys:
@@ -111,6 +134,7 @@ def deduplicate_dataframe(
     count_col = uuid4().hex  # generate a random column name that is virtually certain to not be in the dataset
     window = Window.partitionBy(keys)
 
+    # noinspection PyTypeChecker
     dupes = df.withColumn(count_col, F.count('*').over(window)).filter(F.col(count_col) > 1).drop(count_col)
     if tiebreaking_columns and not df.isStreaming:
         row_number_col = uuid4().hex
@@ -133,8 +157,14 @@ def hash_fields(df: DataFrame, denylist_fields: list[str] = None, algorithm: str
 
     :param df: Input dataframe that is to be hashed.
     :param denylist_fields: Fields that will not be hashed.
-    :param algorithm: The function that is used to generate the hash digest. Includes sha1, sha2, md5, hash, xxhash64.
-    :param num_bits: For SHA2 only. The number of output bits.
+    :param algorithm: The function that is used to generate the hash digest, includes:
+
+            * ``xxhash64`` (default) :class:`pyspark.sql.functions.xxhash64`
+            * ``md5`` :class:`pyspark.sql.functions.md5`
+            * ``sha1`` :class:`pyspark.sql.functions.sha1`
+            * ``sha2`` :class:`pyspark.sql.functions.sha2`
+            * ``hash`` :class:`pyspark.sql.functions.hash`
+    :param num_bits: Only for sha2. The desired bit length of the result.
     :return: A column that represents the hash.
     """
     supported_algorithms = ['sha1', 'sha2', 'md5', 'hash', 'xxhash64']
@@ -178,7 +208,7 @@ def hash_schema(df: DataFrame, denylist_fields: list[str] = None) -> Column:
 
     fields_set = set(all_fields)
     if len(all_fields) != len(fields_set):
-        dupes = [item for item, count in collections.Counter(all_fields).items() if count > 1]
+        dupes = [item for item, count in Counter(all_fields).items() if count > 1]
         raise ValueError(f'Duplicate field(s) detected in df, {dupes}')
 
     """
@@ -190,48 +220,164 @@ def hash_schema(df: DataFrame, denylist_fields: list[str] = None) -> Column:
     return hash_col
 
 
-def _map_fields(df: DataFrame, fields_to_map: list[str], function: Callable) -> DataFrame:
-
-    for field in fields_to_map:
-        df = df.withColumn(field, function(field))
-    return df
-
-
-def map_fields_by_regex(df: DataFrame, regex: str, function: Callable):
-    """
-
-    :param df:
-    :param regex:
-    :param function:
-    :return:
-    """
+def _get_fields_by_regex(df: DataFrame, regex: str) -> list[str]:
     # ChatGPT 🤖 prompt:
     # i have a regex pattern string. write a python program that iterates through a list of strings and returns elements that match the regex
     regex = re.compile(regex)
     all_fields = fields(df)
     matches = [field for field in all_fields if regex.search(field)]
+    return matches
+
+
+def _get_fields_by_type(df: DataFrame, target_type: DataType) -> list[str]:
+    all_fields = fields_with_types(df)
+    print(all_fields)
+    pertinent_fields = [field[0] for field in all_fields if field[1] == target_type]
+    return pertinent_fields
+
+
+def _map_fields(df: DataFrame, fields_to_map: list[str], function: Callable) -> DataFrame:
+    for field in fields_to_map:
+        df = df.withColumn(field, function(field))
+    return df
+
+
+def map_fields_by_regex(df: DataFrame, regex: str, function: Callable) -> DataFrame:
+    """
+
+    Apply a function `function` over fields that match a regular expression.
+
+    :param df:
+    :param regex: Regular expression pattern. Uses Python's `re` module.
+    :param function: Any `pyspark.sql.function` or lambda function that takes a column.
+    :return: Resulting DataFrame
+    """
+
+    matches = _get_fields_by_regex(df, regex)
     return _map_fields(df, matches, function)
 
 
-def map_fields_by_type(df: DataFrame, target_type: DataType, function: Callable):
+def map_fields_by_type(df: DataFrame, target_type: DataType, function: Callable) -> DataFrame:
+    """
+
+    Apply a function `function` over fields that have a target type.
+
+    :param df:
+    :param target_type:
+    :param function: Any `pyspark.sql.function` or lambda function that takes a column.
+    :return: Resulting DataFrame
+    """
+    pertinent_fields = _get_fields_by_type(df, target_type)
+    return _map_fields(df, pertinent_fields, function)
+
+
+def map_fields(df: DataFrame, field_list: list[str], function: Callable) -> DataFrame:
+    """
+
+    Apply a function `function` over fields that are specified in a list.
+
+    :param df:
+    :param field_list:
+    :param function: Any `pyspark.sql.function` or lambda function that takes a column.
+    :return:
+    """
+    return _map_fields(df, field_list, function)
+
+
+def select_fields_by_type(df: DataFrame, target_type: DataType):
     """
 
     :param df:
     :param target_type:
-    :param function:
     :return:
     """
-    all_fields = fields_with_types(df)
-    pertinent_fields = [field[0] for field in all_fields if field[1] == target_type]
-    return _map_fields(df, pertinent_fields, function)
+    pertinent_fields = _get_fields_by_type(df, target_type)
+    return df.select(*pertinent_fields)
 
 
-def map_fields(df: DataFrame, field_list: list[str], function: F):
+def select_fields_by_regex(df: DataFrame, regex: str) -> DataFrame:
     """
 
     :param df:
-    :param field_list:
-    :param function:
+    :param regex:
     :return:
     """
-    return _map_fields(df, field_list, function)
+    matches = _get_fields_by_regex(df, regex)
+    return df.select(*matches)
+
+
+def _drop_fields(fields_to_drop: tuple[str, list[str | None]]) -> tuple[str, Column]:
+
+    address, leaves = fields_to_drop
+
+    if not leaves or leaves[0] is None:
+        raise ValueError(f'Cannot drop top-level field `{address}` with this function. Use df.drop() instead.')
+
+    def _traverse_nest(nest, l, c=0):
+        if len(l) == 0:  # termination condition
+            return F.col(nest).dropFields(*leaves)
+        else:  # recursive step
+            current_level = l[0]
+            return F.col(nest).withField(current_level, _traverse_nest(f'{nest}.{current_level}', l[1:], c + 1))
+
+    levels = address.split('.')
+    if len(levels) == 1:
+        return address, F.col(address).dropFields(*leaves)
+    col = _traverse_nest(levels[0], levels[1:])
+    return levels[0], col
+
+
+def drop_fields(df: DataFrame, fields_to_drop: list[str]) -> DataFrame:
+    """
+
+    Drops a DataFrame's fields, including nested fields and top-level columns.
+
+    :param df:
+    :param fields_to_drop: A list of field names that are to be dropped
+    :return:
+    """
+    # potential optimization, use a trie and resolve trie leafs together, as dropFields takes multiple field args
+
+    tries = _field_trie(fields_to_drop)
+    for trie in tries.items():
+        if trie[1] == [None]:
+            df = df.drop(trie[0])
+        else:
+            name, col = _drop_fields(trie)
+            df = df.withColumn(name, col)
+    return df
+
+
+def infer_json_field(df: DataFrame, target_field: str, options: dict[str, str] = None) -> StructType:
+    """
+
+    Parses a JSON string and infers its schema.
+
+    :param df:
+    :param target_field: A field that contains CSV strings that are to be inferred.
+    :param options: Standard csv reader options, including `header`. See :class:pyspark.sql.DataFrameReader.json
+    :return: The inferred StructType
+    """
+    if not options:  # pragma: no cover
+        options = dict()
+    spark = df.sparkSession
+    rdd = df.select(target_field).rdd.map(lambda row: row[0])  # pragma: no cover
+    return spark.read.options(**options).json(rdd).schema
+
+
+def infer_csv_field(df: DataFrame, target_field: str, options: dict[str, str] = None) -> StructType:
+    """
+
+    Parses a CSV string and infers its schema.
+
+    :param df:
+    :param target_field: A field that contains CSV strings that are to be inferred.
+    :param options: Standard csv reader options, including `header`. See :class:pyspark.sql.DataFrameReader.csv
+    :return: The inferred StructType
+    """
+    if not options:  # pragma: no cover
+        options = dict()
+    spark = df.sparkSession
+    rdd = df.select(target_field).rdd.map(lambda row: row[0])  # pragma: no cover
+    # noinspection PyTypeChecker
+    return spark.read.options(**options).csv(rdd).schema
